#ifdef NETAPP  //BURT350945 make strings MT safe

#include <cstdlib>
#include <cstdio>
#include <cctype>
#include <ext/concurrence.h>
#include <ext/algorithm>
#include <unistd.h>
#include <limits.h>
#include <time.h>
#include <sys/sysctl.h>
#include <machine/cpufunc.h>

// allow to collect the stats for debugging of str_lock()
//#define DEBUG_STR_LOCK_STAT

namespace std
{
    /// The number of mutexes in __assign_mutex[] below.
    /// 257 seems like a reasonable balance, typically this number should be
    /// more than a number of threads that want to work under a 
    /// string lock at the same time
    static const unsigned __NUM_ASSIGN_MUTEXES = 257;

    /// Hash like array of mutexes.
    /// We want to minimize the probability that 2(or more) threads are trying
    /// to lock the same mutex for actually different strings
    static __gthread_mutex_t __assign_mutex[__NUM_ASSIGN_MUTEXES]; // static data are going to be initialized with NULL

/// everything that is enclosed in #ifdef ... #endif is only for debugging
/// of str_lock() and should not be compiled in the build
#ifdef DEBUG_STR_LOCK_STAT

    #define debugStrLock(statement_or_expr) statement_or_expr

    /// true if str_lock() as allowed to collect statistics.
    /// if true str_lock() works many times slower!!!
    static bool __str_stat_enabled = false;

    /// stats with counters that show how many times each mutex was locked
    static uint64_t volatile __str_hash_lock_count[__NUM_ASSIGN_MUTEXES]; // initially zerroed
    // how many times the mutex has been locked right before we call xyz_mutex_lock()
    static uint64_t volatile __str_hash_was_locked_count[__NUM_ASSIGN_MUTEXES]; // initially zerroed

    /// index of the latency bin in __str_lock_*latency that are defined below
    enum ELatencyBinId {
        FROM_ZERO_TO_10NS,
        FROM_10NS_TO_100NS,
        FROM_100NS_TO_1US,
        FROM_1US_TO_10US,
        FROM_10US_TO_100US,
        FROM_100US_TO_1MS,
        FROM_1MS_TO_10MS,
        FROM_10MS_TO_100MS,
        FROM_100MS_TO_1S,
        FROM_1S_TO_10S,
        FROM_10S,
        LATENCY_BIN_ID_COUNT
    };
    /// description of the latency bin that corresponds to the index in ELatencyBinId above
    static const char* latency_bin_description[] = {
        "[0;10ns)",
        "[10ns;100ns)",
        "[100ns;1us)",
        "[1us;10us)",
        "[10us;100us)",
        "[100us;1ms)",
        "[1ms;10ms)",
        "[10ms;100ms)",
        "[100ms;1s)",
        "[1s;10s)",
        "[10s;inf)"
    };
    /// latency of str_lock()
    static uint64_t volatile __str_lock_latency[LATENCY_BIN_ID_COUNT]; // initially zeroed
    /// latency of those str_lock() which likely had to wait because the mutex was locked
    static uint64_t volatile __str_was_locked_latency[LATENCY_BIN_ID_COUNT]; // initially zeroed
    /// latency of str_lock() which locked 2 mutexes and the second one was initially locked
    /// by someone else, but lock succeeded in opposite direction: from second to first
    static uint64_t volatile __str_lock_single_trouble_latency[LATENCY_BIN_ID_COUNT]; // initially zeroed
    /// latency of str_lock() which locked 2 mutexes and the second one was initially locked
    /// by someone else, and lock failed in opposite direction: from second to first;
    /// so there was a spin to obtain both locks
    static uint64_t volatile __str_lock_double_trouble_latency[LATENCY_BIN_ID_COUNT]; // initially zeroed
    /// total latency of all str_lock()
    static uint64_t volatile __str_lock_total_latency = 0;
    /// a constant to translate rdtsc() to nanoseconds
    static double __cycles_per_nsec = 0;

    /// obtain __cycles_per_nsec
    static void configure_cycle_count(void) {
        uint64_t clockspeed = 1000000000;
        size_t len = sizeof( clockspeed );
        sysctlbyname("machdep.tsc_freq", &clockspeed, &len, 0, 0);
        __cycles_per_nsec = clockspeed/1000000000.0;
    }

    /// get current time in nanoseconds
    static uint64_t get_time_ns()
    {
        if (__cycles_per_nsec == 0) {
            configure_cycle_count();
        }
        // rdtsc() is very slow but clock_gettime()
        // is way slower :-(
        uint64_t raw = rdtsc();
        return (uint64_t)(raw/__cycles_per_nsec);
    }
    /// increment the counter in __str_lock_*latency in the element/bin that
    /// correspond(s) to the latency t
    /// @param trouble_count is the number of iterations to lock 2 mutexes,
    ///                      always 0 - if only 1 mutex was locked,
    ///                      it's only used for __str_lock_*_trouble_latency arrays
    static void update_latency_stat(uint64_t t, unsigned trouble_count, bool was_locked)
    {
        (void)__sync_fetch_and_add(&__str_lock_total_latency, t);
        //
        ELatencyBinId bin_id;
        // kinda binary search
        if (t < 10000) { // <10us
            if (t < 100) { // <100ns
                if (t < 10) { // <10ns
                    bin_id = FROM_ZERO_TO_10NS;
                } else {
                    bin_id = FROM_10NS_TO_100NS;
                }
            } else {
                if (t < 1000) {// <1000ns
                    bin_id = FROM_100NS_TO_1US;
                } else {
                    bin_id = FROM_1US_TO_10US;
                }
            }
        } else {
            if (t < 10000000) { // <10ms
                if (t < 100000) { // <100us
                    bin_id = FROM_10US_TO_100US;
                } else if (t < 1000000) { // <1ms
                    bin_id = FROM_100US_TO_1MS;
                } else {
                    bin_id = FROM_1MS_TO_10MS;
                }
            } else {
                if (t < 100000000) { // 100ms
                    bin_id = FROM_10MS_TO_100MS;
                } else if (t < 1000000000) { // 1s
                    bin_id = FROM_100MS_TO_1S;
                } else if (t < 10000000000){ // 10s
                    bin_id = FROM_1S_TO_10S;
                } else {
                    bin_id = FROM_10S;
                }
            }
        }
        (void)__sync_fetch_and_add(&__str_lock_latency[(int)bin_id], 1);
        if (was_locked) {
            (void)__sync_fetch_and_add(&__str_was_locked_latency[(int)bin_id], 1);
        }
        if (trouble_count > 1) {
            (void)__sync_fetch_and_add(&__str_lock_double_trouble_latency[(int)bin_id], 1);
        } else if (trouble_count > 0) {
            (void)__sync_fetch_and_add(&__str_lock_single_trouble_latency[(int)bin_id], 1);
        }
    }
#else

#define debugStrLock(statement_or_expr) ((void)0)

#endif

    /// convert a pointer to a hash value suitable for the index in
    /// the array with mutexes: __assign_mutex[]
    static inline unsigned ptr_to_hash(const void *addr)
    {
        return (((unsigned int)(unsigned long long)addr) >> 4) % __NUM_ASSIGN_MUTEXES;
    }

    /// lock a mutex with index i.
    /// @param i is the index of the mutex in the array __assign_mutex[]
    /// @param was_locked_ptr is only for debugging, can be NULL
    static inline void str_lock_index(unsigned i, bool *was_locked_ptr)
    {
#ifdef DEBUG_STR_LOCK_STAT
        if (was_locked_ptr != NULL && __str_stat_enabled) {
            if (__gthread_mutex_trylock(&__assign_mutex[i]) != 0) {
                *was_locked_ptr = true;
            } else {
                return;
            }
        }
#endif
        __gthread_mutex_lock(&__assign_mutex[i]);
    }

    /// Lock mutex(es) corresponding to the given addr and optionaly addr2.
    /// @param addr is the first address address to lock
    /// @param addr2 is the second address address to lock or NULL if there is
    ///              no second address
    /// @param OUT locked is a pointer to an opaque value which is actually an index of
    ///                   the mutex in __assign_mutex[] that correspons to addr 
    ///                   that has been locked by str_lock()
    /// @param OUT locked2 is a pointer to an opaque value which is actually an index of
    ///                   the mutex in __assign_mutex[] that correspons to addr2 
    ///                   that has been locked by str_lock()
    /// @return always 0
    /// @note *locked and *locked2 should be passed later to str_unlock()
    int str_lock(const void * const volatile * addr, const void * const volatile * addr2,
                 unsigned * locked, unsigned * locked2)
    {
        const void * const volatile * first_addr  = addr;
        const void * const volatile * second_addr = addr2;
        bool *was_locked_ptr = NULL;

#ifdef DEBUG_STR_LOCK_STAT
        uint64_t start_ns;
        unsigned step = 0;
        bool was_locked = false;
        if (__str_stat_enabled) {
            was_locked_ptr  = &was_locked;
            start_ns = get_time_ns();
        }
#endif
        /// We may need to lock 2 mutex.
        /// To prevent a deadlock the mutexes must be locked at the same sequence.
        /// We lock the mutex with the lowest index first.
        for(; ; debugStrLock(step++)) {
            const void *first_a = *first_addr;
            unsigned first_i = ptr_to_hash(first_a);

            const void *second_a = second_addr != NULL ? *second_addr : NULL;
            unsigned second_i = second_addr != NULL ? ptr_to_hash(second_a) : UINT_MAX;

            if (second_i < first_i) { // swap second with first
                std::swap(first_i, second_i);
                std::swap(first_a, second_a);
                std::swap(first_addr, second_addr);
            }

            str_lock_index(first_i, was_locked_ptr);
            // if *first_addr has been changed from another thread then try again
            if (*first_addr != first_a) {
                __gthread_mutex_unlock(&__assign_mutex[first_i]);
                continue;
            }
            // if we need to lock another mutex
            if (second_i != UINT_MAX) {
                second_a = *second_addr; // by this time *second_addr might have changed
                second_i = ptr_to_hash(second_a);
                if (second_i != first_i) {
                    bool lock_both_again;
                    // try to lock the *second_addr
                    while (true) {
                        lock_both_again = false;
                        if (second_i < first_i) {
                            // we are out of luck, need to start from the beginning
                            __gthread_mutex_unlock(&__assign_mutex[first_i]);
                            lock_both_again = true;
                            break;
                        } else if (second_i == first_i) {
                            break;
                        } else {
                            str_lock_index(second_i, was_locked_ptr);
                            // if *second_addr has been changed from another thread then try again
                            if (*second_addr != second_a) {
                                __gthread_mutex_unlock(&__assign_mutex[second_i]);
                                debugStrLock(step++); // the degree of the inefficiency
                                second_a = *second_addr;
                                second_i = ptr_to_hash(second_a);
                            } else {
                                break;
                            }
                        }
                    }
                    if (lock_both_again) {
                        continue;
                    }
                }
            }
            // we are done
            if (first_addr == addr) {
                *locked  = first_i;
                *locked2 = second_i;
            } else {
                *locked  = second_i;
                *locked2 = first_i;
            }
            break;
        }
#ifdef DEBUG_STR_LOCK_STAT
        if (__str_stat_enabled) {
            update_latency_stat(get_time_ns() - start_ns, step, was_locked);
            // update hash stat in __str_hash_lock_count
            (void)__sync_fetch_and_add(&__str_hash_lock_count[*locked], 1);
            if (was_locked) {
                (void)__sync_fetch_and_add(&__str_hash_was_locked_count[*locked], 1);
            }
            if (*locked2 < __NUM_ASSIGN_MUTEXES) {
                (void)__sync_fetch_and_add(&__str_hash_lock_count[*locked2], 1);       
            }
        }
#endif
        return 0;
    }

    /// unlock the mutex(es) locked previously by str_lock(),
    /// see str_lock() above
    int str_unlock(unsigned locked, unsigned locked2)
    {
	int rv = __gthread_mutex_unlock(&__assign_mutex[locked]);
        if (locked2 < __NUM_ASSIGN_MUTEXES && locked2 != locked) {
            __gthread_mutex_unlock(&__assign_mutex[locked2]);
        }
        return rv;
    }
    
    /// enable/disable str_lock() to collect statistics: latency and usage of the mutex hash
    void str_lock_enable_stat(bool enable)
    {
#ifdef DEBUG_STR_LOCK_STAT
        __str_stat_enabled = enable;
#endif
    }

    /// print the statistics (latency and usage of the mutex hash) collected by
    /// str_lock(). 
    void str_lock_stat(char* latency_buf, unsigned int latency_buf_size,
                       char* hash_buf,    unsigned int hash_buf_size)
    {
        if (latency_buf) latency_buf[0] = '\000';
        if (hash_buf) hash_buf[0] = '\000';
#ifdef DEBUG_STR_LOCK_STAT
        if (latency_buf != NULL) {
            uint64_t total_count = 0, was_locked_count = 0, single_trouble_count = 0,
                     double_trouble_count = 0;
            for (int i = 0; i < LATENCY_BIN_ID_COUNT; i++) {
                total_count          += __str_lock_latency[i];
                was_locked_count     += __str_was_locked_latency[i];
                single_trouble_count += __str_lock_single_trouble_latency[i];
                double_trouble_count += __str_lock_double_trouble_latency[i];
                if (latency_buf_size > 0) {
                    int n = snprintf(latency_buf, latency_buf_size, "%-12s: %9llu, try_lock() failed: %9llu, small trouble: %7llu, double trouble: %7llu\n",
                                     latency_bin_description[i],
                                     (unsigned long long)__str_lock_latency[i],
                                     (unsigned long long)__str_was_locked_latency[i],
                                     (unsigned long long)__str_lock_single_trouble_latency[i],
                                     (unsigned long long)__str_lock_double_trouble_latency[i]);
                    if (n <= 0 || (unsigned)n >= latency_buf_size) {
                        latency_buf_size = 0;
                        continue;
                    }
                    latency_buf += n;
                    latency_buf_size -= n;
                }
            }
            
            int n = snprintf(latency_buf, latency_buf_size, "\n%-12s: %9llu, try_lock() failed: %9llu, small trouble: %7llu, double trouble: %7llu\n",
                             "Total",
                             (unsigned long long)total_count,
                             (unsigned long long)was_locked_count,
                             (unsigned long long)single_trouble_count,
                             (unsigned long long)double_trouble_count);
            if (n <= 0 || (unsigned)n >= latency_buf_size) {
                latency_buf_size = 0;
            } else {
                latency_buf += n;
                latency_buf_size -= n;
            }
            if (total_count == 0) {
                total_count = 1;
            }
            snprintf(latency_buf, latency_buf_size, "%-12s: %9llu, try_lock() failed: %8.4g%%, small trouble: %6.4g%%, double trouble: %7.4g%%, cycles_per_nsec=%g\n",
                     "Avg/%",
                     (unsigned long long)__str_lock_total_latency/total_count,
                     (double)was_locked_count/total_count*100,
                     (double)single_trouble_count/total_count*100,
                     (double)double_trouble_count/total_count*100,
                     __cycles_per_nsec);
        }
        if (hash_buf != NULL) {
            uint64_t total_count = 0, was_locked_count = 0;
            unsigned hits99 = 0;
            for (int i = 0; i < __NUM_ASSIGN_MUTEXES; i++) {
                total_count      += __str_hash_lock_count[i];
                was_locked_count += __str_hash_was_locked_count[i];
                if (__str_hash_lock_count[i] > 99) {
                    hits99++;
                }
                if (hash_buf_size > 0) {
                    int n = snprintf(hash_buf, hash_buf_size, "%3d: %8llu, try_lock() failed: %6llu\n",
                                     i, (unsigned long long)__str_hash_lock_count[i],
                                     (unsigned long long)__str_hash_was_locked_count[i]);
                    if (n <= 0 || (unsigned)n >= hash_buf_size) {
                        hash_buf_size = 0;
                        continue;
                    }
                    hash_buf += n;
                    hash_buf_size -= n;
                }
            }
            int n = snprintf(hash_buf, hash_buf_size, "\nTot: %8llu, try_lock() failed: %6llu, bins with more than 99 hits: %u\n",
                             (unsigned long long)total_count,
                             (unsigned long long)was_locked_count, hits99);
            if (n <= 0 || (unsigned)n >= hash_buf_size) {
                hash_buf_size = 0;
            } else {
                hash_buf += n;
                hash_buf_size -= n;
            }
            snprintf(hash_buf, hash_buf_size, "Avg: %-8s, try_lock() failed: %6.5g%%\n",
                     "", (double)was_locked_count/total_count*100);
        }
#endif
    }
}

#endif // NETAPP
